// ExtJS related things

 // do not send '_dc' parameter
Ext.Ajax.disableCaching = false;

// custom Vtypes
Ext.apply(Ext.form.field.VTypes, {
    IPAddress: function(v) {
	return Proxmox.Utils.IP4_match.test(v);
    },
    IPAddressText: gettext('Example') + ': 192.168.1.1',
    IPAddressMask: /[\d.]/i,

    IPCIDRAddress: function(v) {
	let result = Proxmox.Utils.IP4_cidr_match.exec(v);
	// limits according to JSON Schema see
	// pve-common/src/PVE/JSONSchema.pm
	return result !== null && result[1] >= 8 && result[1] <= 32;
    },
    IPCIDRAddressText: gettext('Example') + ': 192.168.1.1/24<br>' + gettext('Valid CIDR Range') + ': 8-32',
    IPCIDRAddressMask: /[\d./]/i,

    IP6Address: function(v) {
        return Proxmox.Utils.IP6_match.test(v);
    },
    IP6AddressText: gettext('Example') + ': 2001:DB8::42',
    IP6AddressMask: /[A-Fa-f0-9:]/,

    IP6CIDRAddress: function(v) {
	let result = Proxmox.Utils.IP6_cidr_match.exec(v);
	// limits according to JSON Schema see
	// pve-common/src/PVE/JSONSchema.pm
	return result !== null && result[1] >= 8 && result[1] <= 128;
    },
    IP6CIDRAddressText: gettext('Example') + ': 2001:DB8::42/64<br>' + gettext('Valid CIDR Range') + ': 8-128',
    IP6CIDRAddressMask: /[A-Fa-f0-9:/]/,

    IP6PrefixLength: function(v) {
	return v >= 0 && v <= 128;
    },
    IP6PrefixLengthText: gettext('Example') + ': X, where 0 <= X <= 128',
    IP6PrefixLengthMask: /[0-9]/,

    IP64Address: function(v) {
        return Proxmox.Utils.IP64_match.test(v);
    },
    IP64AddressText: gettext('Example') + ': 192.168.1.1 2001:DB8::42',
    IP64AddressMask: /[A-Fa-f0-9.:]/,

    IP64CIDRAddress: function(v) {
	let result = Proxmox.Utils.IP64_cidr_match.exec(v);
	if (result === null) {
	    return false;
	}
	if (result[1] !== undefined) {
	    return result[1] >= 8 && result[1] <= 128;
	} else if (result[2] !== undefined) {
	    return result[2] >= 8 && result[2] <= 32;
	} else {
	    return false;
	}
    },
    IP64CIDRAddressText: gettext('Example') + ': 192.168.1.1/24 2001:DB8::42/64',
    IP64CIDRAddressMask: /[A-Fa-f0-9.:/]/,

    MacAddress: function(v) {
	return (/^([a-fA-F0-9]{2}:){5}[a-fA-F0-9]{2}$/).test(v);
    },
    MacAddressMask: /[a-fA-F0-9:]/,
    MacAddressText: gettext('Example') + ': 01:23:45:67:89:ab',

    MacPrefix: function(v) {
	return (/^[a-f0-9][02468ace](?::[a-f0-9]{2}){0,2}:?$/i).test(v);
    },
    MacPrefixMask: /[a-fA-F0-9:]/,
    MacPrefixText: gettext('Example') + ': 02:8f - ' + gettext('only unicast addresses are allowed'),

    BridgeName: function(v) {
        return (/^vmbr\d{1,4}$/).test(v);
    },
    VlanName: function(v) {
       if (Proxmox.Utils.VlanInterface_match.test(v)) {
         return true;
       } else if (Proxmox.Utils.Vlan_match.test(v)) {
         return true;
       }
       return true;
    },
    BridgeNameText: gettext('Format') + ': vmbr<b>N</b>, where 0 <= <b>N</b> <= 9999',

    BondName: function(v) {
        return (/^bond\d{1,4}$/).test(v);
    },
    BondNameText: gettext('Format') + ': bond<b>N</b>, where 0 <= <b>N</b> <= 9999',

    InterfaceName: function(v) {
        return (/^[a-z][a-z0-9_]{1,20}$/).test(v);
    },
    InterfaceNameText: gettext("Allowed characters") + ": 'a-z', '0-9', '_'<br />" +
		       gettext("Minimum characters") + ": 2<br />" +
		       gettext("Maximum characters") + ": 21<br />" +
		       gettext("Must start with") + ": 'a-z'",

    StorageId: function(v) {
        return (/^[a-z][a-z0-9\-_.]*[a-z0-9]$/i).test(v);
    },
    StorageIdText: gettext("Allowed characters") + ":  'A-Z', 'a-z', '0-9', '-', '_', '.'<br />" +
		   gettext("Minimum characters") + ": 2<br />" +
		   gettext("Must start with") + ": 'A-Z', 'a-z'<br />" +
		   gettext("Must end with") + ": 'A-Z', 'a-z', '0-9'<br />",

    ConfigId: function(v) {
        return (/^[a-z][a-z0-9_]+$/i).test(v);
    },
    ConfigIdText: gettext("Allowed characters") + ": 'A-Z', 'a-z', '0-9', '_'<br />" +
		  gettext("Minimum characters") + ": 2<br />" +
		  gettext("Must start with") + ": " + gettext("letter"),

    HttpProxy: function(v) {
        return (/^http:\/\/.*$/).test(v);
    },
    HttpProxyText: gettext('Example') + ": http://username:password&#64;host:port/",

    DnsName: function(v) {
	return Proxmox.Utils.DnsName_match.test(v);
    },
    DnsNameText: gettext('This is not a valid DNS name'),

    DnsNameOrWildcard: function(v) {
	return Proxmox.Utils.DnsName_or_Wildcard_match.test(v);
    },
    DnsNameOrWildcardText: gettext('This is not a valid DNS name'),

    // workaround for https://www.sencha.com/forum/showthread.php?302150
    proxmoxMail: function(v) {
        return (/^(\w+)([-+.][\w]+)*@(\w[-\w]*\.){1,5}([A-Za-z]){2,63}$/).test(v);
    },
    proxmoxMailText: gettext('Example') + ": user@example.com",

    DnsOrIp: function(v) {
	if (!Proxmox.Utils.DnsName_match.test(v) &&
	    !Proxmox.Utils.IP64_match.test(v)) {
	    return false;
	}

	return true;
    },
    DnsOrIpText: gettext('Not a valid DNS name or IP address.'),

    HostPort: function(v) {
	return Proxmox.Utils.HostPort_match.test(v) ||
		Proxmox.Utils.HostPortBrackets_match.test(v) ||
		Proxmox.Utils.IP6_dotnotation_match.test(v);
    },
    HostPortText: gettext('Host/IP address or optional port is invalid'),

    HostList: function(v) {
	let list = v.split(/[ ,;]+/);
	let i;
	for (i = 0; i < list.length; i++) {
	    if (list[i] === '') {
		continue;
	    }

	    if (!Proxmox.Utils.HostPort_match.test(list[i]) &&
		!Proxmox.Utils.HostPortBrackets_match.test(list[i]) &&
		!Proxmox.Utils.IP6_dotnotation_match.test(list[i])) {
		return false;
	    }
	}

	return true;
    },
    HostListText: gettext('Not a valid list of hosts'),

    password: function(val, field) {
        if (field.initialPassField) {
            let pwd = field.up('form').down(`[name=${field.initialPassField}]`);
            return val === pwd.getValue();
        }
        return true;
    },

    passwordText: gettext('Passwords do not match'),

    email: function(value) {
	let emailre = /^[\w+~-]+(\.[\w+~-]+)*@[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$/;
	return emailre.test(value);
    },
});

// we always want the number in x.y format and never in, e.g., x,y
Ext.define('PVE.form.field.Number', {
    override: 'Ext.form.field.Number',
    submitLocaleSeparator: false,
});

// ExtJs 5-6 has an issue with caching
// see https://www.sencha.com/forum/showthread.php?308989
Ext.define('Proxmox.UnderlayPool', {
    override: 'Ext.dom.UnderlayPool',

    checkOut: function() {
        let cache = this.cache,
            len = cache.length,
            el;

        // do cleanup because some of the objects might have been destroyed
	while (len--) {
            if (cache[len].destroyed) {
                cache.splice(len, 1);
            }
        }
        // end do cleanup

	el = cache.shift();

        if (!el) {
            el = Ext.Element.create(this.elementConfig);
            el.setVisibilityMode(2);
            //<debug>
            // tell the spec runner to ignore this element when checking if the dom is clean
	    el.dom.setAttribute('data-sticky', true);
            //</debug>
	}

        return el;
    },
});

// if the order of the values are not the same in originalValue and value
// extjs will not overwrite value, but marks the field dirty and thus
// the reset button will be enabled (but clicking it changes nothing)
// so if the arrays are not the same after resetting, we
// clear and set it
Ext.define('Proxmox.form.ComboBox', {
    override: 'Ext.form.field.ComboBox',

    reset: function() {
	// copied from combobox
	let me = this;
	me.callParent();

	// clear and set when not the same
	let value = me.getValue();
	if (Ext.isArray(me.originalValue) && Ext.isArray(value) &&
	    !Ext.Array.equals(value, me.originalValue)) {
	    me.clearValue();
	    me.setValue(me.originalValue);
	}
    },

    // we also want to open the trigger on editable comboboxes by default
    initComponent: function() {
	let me = this;
	me.callParent();

	if (me.editable) {
	    // The trigger.picker causes first a focus event on the field then
	    // toggles the selection picker. Thus skip expanding in this case,
	    // else our focus listener expands and the picker.trigger then
	    // collapses it directly afterwards.
	    Ext.override(me.triggers.picker, {
		onMouseDown: function(e) {
		    // copied "should we focus" check from Ext.form.trigger.Trigger
		    if (e.pointerType !== 'touch' && !this.field.owns(Ext.Element.getActiveElement())) {
			me.skip_expand_on_focus = true;
		    }
		    this.callParent(arguments);
		},
	    });

	    me.on("focus", function(combobox) {
		if (!combobox.isExpanded && !combobox.skip_expand_on_focus) {
		    combobox.expand();
		}
		combobox.skip_expand_on_focus = false;
	    });
	}
    },
});

// when refreshing a grid/tree view, restoring the focus moves the view back to
// the previously focused item. Save scroll position before refocusing.
Ext.define(null, {
    override: 'Ext.view.Table',

    jumpToFocus: false,

    saveFocusState: function() {
	var me = this,
	    store = me.dataSource,
	    actionableMode = me.actionableMode,
	    navModel = me.getNavigationModel(),
	    focusPosition = actionableMode ? me.actionPosition : navModel.getPosition(true),
	    activeElement = Ext.fly(Ext.Element.getActiveElement()),
	    focusCell = focusPosition && focusPosition.view === me &&
	    Ext.fly(focusPosition.getCell(true)),
	    refocusRow, refocusCol, record;

	// The navModel may return a position that is in a locked partner, so check that
	// the focusPosition's cell contains the focus before going forward.
	// The skipSaveFocusState is set by Actionables which actively control
	// focus destination. See CellEditing#activateCell.
	if (!me.skipSaveFocusState && focusCell && focusCell.contains(activeElement)) {
	    // Separate this from the instance that the nav model is using.
	    focusPosition = focusPosition.clone();

	    // While we deactivate the focused element, suspend focus processing on it.
	    activeElement.suspendFocusEvents();

	    // Suspend actionable mode.
	    // Each Actionable must silently save its state ready to resume when focus
	    // can be restored but should only do that if the activeElement is not the cell itself,
	    // this happens when the grid is refreshed while one of the actionables is being
	    // deactivated (e.g. Calling  view refresh inside CellEditor 'edit' event listener).
	    if (actionableMode && focusCell.dom !== activeElement.dom) {
		me.suspendActionableMode();
	    } else {
		// Clear position, otherwise the setPosition on the other side
		// will be rejected as a no-op if the resumption position is logically
		// equivalent.
		actionableMode = false;
		navModel.setPosition();
	    }

	    // Do not leave the element in tht state in case refresh fails, and restoration
	    // closure not called.
	    activeElement.resumeFocusEvents();

	    // if the store is expanding or collapsing, we should never scroll the view.
	    if (store.isExpandingOrCollapsing) {
		return Ext.emptyFn;
	    }

	    // The following function will attempt to refocus back in the same mode to the same cell
	    // as it was at before based upon the previous record (if it's still in the store),
	    // or the row index.
	    return function() {
		var all;

		// May have changed due to reconfigure
		store = me.dataSource;

		// If we still have data, attempt to refocus in the same mode.
		if (store.getCount()) {
		    all = me.all;

		    // Adjust expectations of where we are able to refocus according to
		    // what kind of destruction might have been wrought on this view's DOM
		    // during focus save.
		    refocusRow =
			Math.min(Math.max(focusPosition.rowIdx, all.startIndex), all.endIndex);

		    refocusCol = Math.min(
			focusPosition.colIdx,
			me.getVisibleColumnManager().getColumns().length - 1,
		    );

		    record = focusPosition.record;

		    focusPosition = new Ext.grid.CellContext(me).setPosition(
			record && store.contains(record) && !record.isCollapsedPlaceholder
			? record
			: refocusRow,
			refocusCol,
		    );

		    // Maybe there are no cells. eg: all groups collapsed.
		    if (focusPosition.getCell(true)) {
			if (actionableMode) {
			    me.resumeActionableMode(focusPosition);
			} else {
			    // we sometimes want to scroll back to where we are

			    let x = me.getScrollX();
			    let y = me.getScrollY();

			    // Pass "preventNavigation" as true
			    // so that that does not cause selection.
			    navModel.setPosition(focusPosition, null, null, null, true);

			    if (!navModel.getPosition()) {
				focusPosition.column.focus();
			    }

			    if (!me.jumpToFocus) {
				me.scrollTo(x, y);
			    }
			}
		    }
		} else { // No rows - focus associated column header
		    focusPosition.column.focus();
		}
	    };
	}
	return Ext.emptyFn;
    },
});

// ExtJS 6.0.1 has no setSubmitValue() (although you find it in the docs).
// Note: this.submitValue is a boolean flag, whereas getSubmitValue() returns
// data to be submitted.
Ext.define('Proxmox.form.field.Text', {
    override: 'Ext.form.field.Text',

    setSubmitValue: function(v) {
	this.submitValue = v;
    },
});

// make mousescrolling work in firefox in the containers overflowhandler,
// by using only the 'wheel' event not 'mousewheel'(fixed in 7.3)
// also reverse the scrolldirection (fixed in 7.3)
// and reduce the default increment
Ext.define(null, {
    override: 'Ext.layout.container.boxOverflow.Scroller',

    wheelIncrement: 1,

    getWheelDelta: function(e) {
	return -e.getWheelDelta(e);
    },

    onOwnerRender: function(owner) {
	var me = this,
	    scrollable = {
		isBoxOverflowScroller: true,
		x: false,
		y: false,
		listeners: {
		    scrollend: this.onScrollEnd,
		    scope: this,
		},
	    };

	// If no obstrusive scrollbars, allow natural scrolling on mobile touch devices
	if (!Ext.scrollbar.width() && !Ext.platformTags.desktop) {
	    scrollable[owner.layout.horizontal ? 'x' : 'y'] = true;
	} else {
	    me.wheelListener = me.layout.innerCt.on(
		'wheel', me.onMouseWheel, me, { destroyable: true },
	    );
	}

	owner.setScrollable(scrollable);
    },
});

// extj 6.7 reversed mousewheel direction... (fixed in 7.3)
// https://forum.sencha.com/forum/showthread.php?472517-Mousewheel-scroll-direction-in-numberfield-with-spinners
// alse use the 'wheel' event instead of 'mousewheel' (fixed in 7.3)
Ext.define('Proxmox.form.field.Spinner', {
    override: 'Ext.form.field.Spinner',

    onRender: function() {
	var me = this,
	    spinnerTrigger = me.getTrigger('spinner');

	me.callParent();

	// Init up/down arrow keys
	if (me.keyNavEnabled) {
	    me.spinnerKeyNav = new Ext.util.KeyNav({
		target: me.inputEl,
		scope: me,
		up: me.spinUp,
		down: me.spinDown,
	    });

	    me.inputEl.on({
		keyup: me.onInputElKeyUp,
		scope: me,
	    });
	}

	// Init mouse wheel
	if (me.mouseWheelEnabled) {
	    me.mon(me.bodyEl, 'wheel', me.onMouseWheel, me);
	}

	// in v4 spinUpEl/spinDownEl were childEls, now they are children of the trigger.
	// create references for compatibility
	me.spinUpEl = spinnerTrigger.upEl;
	me.spinDownEl = spinnerTrigger.downEl;
    },

    onMouseWheel: function(e) {
	var me = this,
	    delta;
	if (me.hasFocus) {
	    delta = e.getWheelDelta();
	    if (delta > 0) {
		me.spinDown();
	    } else if (delta < 0) {
		me.spinUp();
	    }
	    e.stopEvent();
	    me.onSpinEnd();
	}
    },
});

// add '@' to the valid id
Ext.define('Proxmox.validIdReOverride', {
    override: 'Ext.Component',
    validIdRe: /^[a-z_][a-z0-9\-_@]*$/i,
});

Ext.define('Proxmox.selection.CheckboxModel', {
    override: 'Ext.selection.CheckboxModel',

    // [P] use whole checkbox cell to multiselect, not only the checkbox
    checkSelector: '.x-grid-cell-row-checker',

    // TODO: remove all optimizations below to an override for parent 'Ext.selection.Model' ??

    // [ P: optimized to remove all records at once as single remove is O(n^3) slow ]
    // records can be an index, a record or an array of records
    doDeselect: function(records, suppressEvent) {
        var me = this,
            selected = me.selected,
            i = 0,
            len, record,
            commit;
        if (me.locked || !me.store) {
            return false;
        }
        if (typeof records === "number") {
            // No matching record, jump out
            record = me.store.getAt(records);
            if (!record) {
                return false;
            }
            records = [
                record,
            ];
        } else if (!Ext.isArray(records)) {
            records = [
                records,
            ];
        }
	// [P] a beforedeselection, triggered by me.onSelectChange below, can block removal by
	// returning false, thus the original implementation removed only here in the commit fn,
	// which has an abysmal performance O(n^3). As blocking removal is not the norm, go do the
	// reverse, record blocked records and remove them from the to-be-removed array before
	// applying it. A FF86 i9-9900K on 10k records goes from >40s to ~33ms for >90% deselection
	let committed = false;
	commit = function() {
	    committed = true;
	    if (record === me.selectionStart) {
		me.selectionStart = null;
	    }
	};
	let removalBlocked = [];
        len = records.length;
        me.suspendChanges();
        for (; i < len; i++) {
            record = records[i];
            if (me.isSelected(record)) {
		committed = false;
                me.onSelectChange(record, false, suppressEvent, commit);
		if (!committed) {
		    removalBlocked.push(record);
		}
                if (me.destroyed) {
                    return false;
                }
            }
        }
	if (removalBlocked.length > 0) {
	    records.remove(removalBlocked);
	}
	selected.remove(records); // [P] FAST(er)
	me.lastSelected = selected.last();
        me.resumeChanges();
        // fire selchange if there was a change and there is no suppressEvent flag
	me.maybeFireSelectionChange(records.length > 0 && !suppressEvent);
	return records.length;
    },


    doMultiSelect: function(records, keepExisting, suppressEvent) {
        var me = this,
            selected = me.selected,
            change = false,
            result, i, len, record, commit;

        if (me.locked) {
            return;
        }

        records = !Ext.isArray(records) ? [records] : records;
        len = records.length;
        if (!keepExisting && selected.getCount() > 0) {
            result = me.deselectDuringSelect(records, suppressEvent);
            if (me.destroyed) {
                return;
            }
            if (result[0]) {
                // We had a failure during selection, so jump out
                // Fire selection change if we did deselect anything
                me.maybeFireSelectionChange(result[1] > 0 && !suppressEvent);
                return;
            } else {
                // Means something has been deselected, so we've had a change
                change = result[1] > 0;
            }
        }

	let gotBlocked, blockedRecords = [];
        commit = function() {
            if (!selected.getCount()) {
                me.selectionStart = record;
            }
	    gotBlocked = false;
            change = true;
        };

        for (i = 0; i < len; i++) {
            record = records[i];
            if (me.isSelected(record)) {
                continue;
            }

	    gotBlocked = true;
            me.onSelectChange(record, true, suppressEvent, commit);
            if (me.destroyed) {
                return;
            }
	    if (gotBlocked) {
		blockedRecords.push(record);
	    }
        }
	if (blockedRecords.length > 0) {
	    records.remove(blockedRecords);
	}
        selected.add(records);
        me.lastSelected = record;

        // fire selchange if there was a change and there is no suppressEvent flag
        me.maybeFireSelectionChange(change && !suppressEvent);
    },
    deselectDuringSelect: function(toSelect, suppressEvent) {
        var me = this,
            selected = me.selected.getRange(),
            changed = 0,
            failed = false;
        // Prevent selection change events from firing, will happen during select
        me.suspendChanges();
        me.deselectingDuringSelect = true;
	let toDeselect = selected.filter(item => !Ext.Array.contains(toSelect, item));
	if (toDeselect.length > 0) {
	    changed = me.doDeselect(toDeselect, suppressEvent);
	    if (!changed) {
		failed = true;
            }
            if (me.destroyed) {
                failed = true;
                changed = 0;
            }
        }
        me.deselectingDuringSelect = false;
        me.resumeChanges();
        return [
            failed,
            changed,
        ];
    },
});

// stop nulling of properties
Ext.define('Proxmox.Component', {
    override: 'Ext.Component',
    clearPropertiesOnDestroy: false,
});

// force alert boxes to be rendered with an Error Icon
// since Ext.Msg is an object and not a prototype, we need to override it
// after the framework has been initiated
Ext.onReady(function() {
    Ext.override(Ext.Msg, {
	alert: function(title, message, fn, scope) { // eslint-disable-line consistent-return
	    if (Ext.isString(title)) {
		let config = {
		    title: title,
		    message: message,
		    icon: this.ERROR,
		    buttons: this.OK,
		    fn: fn,
		    scope: scope,
		    minWidth: this.minWidth,
		};
	    return this.show(config);
	    }
	},
    });
});
Ext.define('Ext.ux.IFrame', {
    extend: 'Ext.Component',

    alias: 'widget.uxiframe',

    loadMask: 'Loading...',

    src: 'about:blank',

    renderTpl: [
        '<iframe src="{src}" id="{id}-iframeEl" data-ref="iframeEl" name="{frameName}" width="100%" height="100%" frameborder="0" allowfullscreen="true"></iframe>',
    ],
    childEls: ['iframeEl'],

    initComponent: function() {
        this.callParent();

        this.frameName = this.frameName || this.id + '-frame';
    },

    initEvents: function() {
        let me = this;
        me.callParent();
        me.iframeEl.on('load', me.onLoad, me);
    },

    initRenderData: function() {
        return Ext.apply(this.callParent(), {
            src: this.src,
            frameName: this.frameName,
        });
    },

    getBody: function() {
        let doc = this.getDoc();
        return doc.body || doc.documentElement;
    },

    getDoc: function() {
        try {
            return this.getWin().document;
        } catch (ex) {
            return null;
        }
    },

    getWin: function() {
        let me = this,
            name = me.frameName,
            win = Ext.isIE
                ? me.iframeEl.dom.contentWindow
                : window.frames[name];
        return win;
    },

    getFrame: function() {
        let me = this;
        return me.iframeEl.dom;
    },

    beforeDestroy: function() {
        this.cleanupListeners(true);
        this.callParent();
    },

    cleanupListeners: function(destroying) {
        let doc, prop;

        if (this.rendered) {
            try {
                doc = this.getDoc();
                if (doc) {
                    Ext.get(doc).un(this._docListeners);
                    if (destroying && doc.hasOwnProperty) {
                        for (prop in doc) {
                            if (Object.prototype.hasOwnProperty.call(doc, prop)) {
                                delete doc[prop];
                            }
                        }
                    }
                }
            } catch (e) {
                // do nothing
            }
        }
    },

    onLoad: function() {
        let me = this,
            doc = me.getDoc(),
            fn = me.onRelayedEvent;

        if (doc) {
            try {
                // These events need to be relayed from the inner document (where they stop
                // bubbling) up to the outer document. This has to be done at the DOM level so
                // the event reaches listeners on elements like the document body. The effected
                // mechanisms that depend on this bubbling behavior are listed to the right
                // of the event.
                Ext.get(doc).on(
                    me._docListeners = {
                        mousedown: fn, // menu dismisal (MenuManager) and Window onMouseDown (toFront)
                        mousemove: fn, // window resize drag detection
                        mouseup: fn, // window resize termination
                        click: fn, // not sure, but just to be safe
                        dblclick: fn, // not sure again
                        scope: me,
                    },
                );
            } catch (e) {
                // cannot do this xss
            }

            // We need to be sure we remove all our events from the iframe on unload or we're going to LEAK!
            Ext.get(this.getWin()).on('beforeunload', me.cleanupListeners, me);

            this.el.unmask();
            this.fireEvent('load', this);
        } else if (me.src) {
            this.el.unmask();
            this.fireEvent('error', this);
        }
    },

    onRelayedEvent: function(event) {
        // relay event from the iframe's document to the document that owns the iframe...

        let iframeEl = this.iframeEl,

            // Get the left-based iframe position
            iframeXY = iframeEl.getTrueXY(),
            originalEventXY = event.getXY(),

            // Get the left-based XY position.
            // This is because the consumer of the injected event will
            // perform its own RTL normalization.
            eventXY = event.getTrueXY();

        // the event from the inner document has XY relative to that document's origin,
        // so adjust it to use the origin of the iframe in the outer document:
        event.xy = [iframeXY[0] + eventXY[0], iframeXY[1] + eventXY[1]];

        event.injectEvent(iframeEl); // blame the iframe for the event...

        event.xy = originalEventXY; // restore the original XY (just for safety)
    },

    load: function(src) {
        let me = this,
            text = me.loadMask,
            frame = me.getFrame();

        if (me.fireEvent('beforeload', me, src) !== false) {
            if (text && me.el) {
                me.el.mask(text);
            }

            frame.src = me.src = src || me.src;
        }
    },
});
// Code39 barcode generator
// see https://en.wikipedia.org/wiki/Code_39

// IBM LTO Ultrium Cartridge Label Specification
// http://www-01.ibm.com/support/docview.wss?uid=ssg1S7000429

const code39_codes = {
    "1": ['B', 's', 'b', 'S', 'b', 's', 'b', 's', 'B'],
    "A": ['B', 's', 'b', 's', 'b', 'S', 'b', 's', 'B'],
    "K": ['B', 's', 'b', 's', 'b', 's', 'b', 'S', 'B'],
    "U": ['B', 'S', 'b', 's', 'b', 's', 'b', 's', 'B'],

    "2": ['b', 's', 'B', 'S', 'b', 's', 'b', 's', 'B'],
    "B": ['b', 's', 'B', 's', 'b', 'S', 'b', 's', 'B'],
    "L": ['b', 's', 'B', 's', 'b', 's', 'b', 'S', 'B'],
    "V": ['b', 'S', 'B', 's', 'b', 's', 'b', 's', 'B'],

    "3": ['B', 's', 'B', 'S', 'b', 's', 'b', 's', 'b'],
    "C": ['B', 's', 'B', 's', 'b', 'S', 'b', 's', 'b'],
    "M": ['B', 's', 'B', 's', 'b', 's', 'b', 'S', 'b'],
    "W": ['B', 'S', 'B', 's', 'b', 's', 'b', 's', 'b'],

    "4": ['b', 's', 'b', 'S', 'B', 's', 'b', 's', 'B'],
    "D": ['b', 's', 'b', 's', 'B', 'S', 'b', 's', 'B'],
    "N": ['b', 's', 'b', 's', 'B', 's', 'b', 'S', 'B'],
    "X": ['b', 'S', 'b', 's', 'B', 's', 'b', 's', 'B'],

    "5": ['B', 's', 'b', 'S', 'B', 's', 'b', 's', 'b'],
    "E": ['B', 's', 'b', 's', 'B', 'S', 'b', 's', 'b'],
    "O": ['B', 's', 'b', 's', 'B', 's', 'b', 'S', 'b'],
    "Y": ['B', 'S', 'b', 's', 'B', 's', 'b', 's', 'b'],

    "6": ['b', 's', 'B', 'S', 'B', 's', 'b', 's', 'b'],
    "F": ['b', 's', 'B', 's', 'B', 'S', 'b', 's', 'b'],
    "P": ['b', 's', 'B', 's', 'B', 's', 'b', 'S', 'b'],
    "Z": ['b', 'S', 'B', 's', 'B', 's', 'b', 's', 'b'],

    "7": ['b', 's', 'b', 'S', 'b', 's', 'B', 's', 'B'],
    "G": ['b', 's', 'b', 's', 'b', 'S', 'B', 's', 'B'],
    "Q": ['b', 's', 'b', 's', 'b', 's', 'B', 'S', 'B'],
    "-": ['b', 'S', 'b', 's', 'b', 's', 'B', 's', 'B'],

    "8": ['B', 's', 'b', 'S', 'b', 's', 'B', 's', 'b'],
    "H": ['B', 's', 'b', 's', 'b', 'S', 'B', 's', 'b'],
    "R": ['B', 's', 'b', 's', 'b', 's', 'B', 'S', 'b'],
    ".": ['B', 'S', 'b', 's', 'b', 's', 'B', 's', 'b'],

    "9": ['b', 's', 'B', 'S', 'b', 's', 'B', 's', 'b'],
    "I": ['b', 's', 'B', 's', 'b', 'S', 'B', 's', 'b'],
    "S": ['b', 's', 'B', 's', 'b', 's', 'B', 'S', 'b'],
    " ": ['b', 'S', 'B', 's', 'b', 's', 'B', 's', 'b'],

    "0": ['b', 's', 'b', 'S', 'B', 's', 'B', 's', 'b'],
    "J": ['b', 's', 'b', 's', 'B', 'S', 'B', 's', 'b'],
    "T": ['b', 's', 'b', 's', 'B', 's', 'B', 'S', 'b'],
    "*": ['b', 'S', 'b', 's', 'B', 's', 'B', 's', 'b'],
};

const colors = [
    '#BB282E',
    '#FAE54A',
    '#9AC653',
    '#01A5E2',
    '#9EAAB6',
    '#D97E35',
    '#E27B99',
    '#67A945',
    '#F6B855',
    '#705A81',
];

const lto_label_width = 70;
const lto_label_height = 16.9;

function foreach_label(page_layout, callback) {
    let count = 0;
    let row = 0;
    let height = page_layout.margin_top;

    while ((height + page_layout.label_height) <= page_layout.page_height) {
	let column = 0;
	let width = page_layout.margin_left;

	while ((width + page_layout.label_width) <= page_layout.page_width) {
	    callback(column, row, count, width, height);
	    count += 1;

	    column += 1;
	    width += page_layout.label_width;
	    width += page_layout.column_spacing;
	}

	row += 1;
	height += page_layout.label_height;
	height += page_layout.row_spacing;
    }
}

function compute_max_labels(page_layout) {
    let max_labels = 0;
    foreach_label(page_layout, function() { max_labels += 1; });
    return max_labels;
}

function svg_label(mode, label, label_type, pagex, pagey, label_borders) {
    let svg = "";

    if (label.length !== 6) {
	throw "wrong label length";
    }
    if (label_type.length !== 2) {
	throw "wrong label_type length";
    }

    let ratio = 2.75;
    let parts = 3*ratio + 6; // 3*wide + 6*small;
    let barcode_width = (lto_label_width/12)*10; // 10*code + 2margin
    let small = barcode_width/(parts*10 + 9);
    let code_width = small*parts;
    let wide = small*ratio;
    let xpos = pagex + code_width;
    let height = 12;

    let label_rect = `x='${pagex}' y='${pagey}' width='${lto_label_width}' height='${lto_label_height}'`;

    if (mode === 'placeholder') {
	if (label_borders) {
	    svg += `<rect class='unprintable' ${label_rect} fill='none' style='stroke:black;stroke-width:0.1;'/>`;
	}
	return svg;
    }
    if (label_borders) {
	svg += `<rect ${label_rect} fill='none' style='stroke:black;stroke-width:0.1;'/>`;
    }

    if (mode === "color" || mode === "frame") {
	let w = lto_label_width/8;
	let h = lto_label_height - height;
	for (let i = 0; i < 7; i++) {
	    let textx = w/2 + pagex + i*w;
	    let texty = pagey;

	    let fill = "none";
	    if (mode === "color" && (i < 6)) {
		let letter = label.charAt(i);
		if (letter >= '0' && letter <= '9') {
		    fill = colors[parseInt(letter, 10)];
		}
	    }

	    svg += `<rect x='${textx}' y='${texty}' width='${w}' height='${h}' style='stroke:black;stroke-width:0.2;fill:${fill};'/>`;

	    if (i == 6) {
		textx += 3;
		texty += 3.7;
		svg += `<text x='${textx}' y='${texty}' style='font-weight:bold;font-size:3px;font-family:sans-serif;'>${label_type}</text>`;
	    } else {
		let letter = label.charAt(i);
		textx += 3.5;
		texty += 4;
		svg += `<text x='${textx}' y='${texty}' style='font-weight:bold;font-size:4px;font-family:sans-serif;'>${letter}</text>`;
	    }
	}
    }

    let raw_label = `*${label}${label_type}*`;

    for (let i = 0; i < raw_label.length; i++) {
	let letter = raw_label.charAt(i);

	let code = code39_codes[letter];
	if (code === undefined) {
	    throw `unable to encode letter '${letter}' with code39`;
	}

	if (mode === "simple") {
	    let textx = xpos + code_width/2;
	    let texty = pagey + 4;

	    if (i > 0 && (i+1) < raw_label.length) {
		svg += `<text x='${textx}' y='${texty}' style='font-weight:bold;font-size:4px;font-family:sans-serif;'>${letter}</text>`;
	    }
	}

	for (let c of code) {
	    if (c === 's') {
		xpos += small;
		continue;
	    }
	    if (c === 'S') {
		xpos += wide;
		continue;
	    }

	    let w = c === 'B' ? wide : small;
	    let ypos = pagey + lto_label_height - height;

	    svg += `<rect x='${xpos}' y='${ypos}' width='${w}' height='${height}' style='fill:black'/>`;
	    xpos = xpos + w;
	}
	xpos += small;
    }

    return svg;
}

function html_page_header() {
    let html = "<html5>";

    html += "<style>";

    /* no page margins */
    html += "@page{margin-left: 0px;margin-right: 0px;margin-top: 0px;margin-bottom: 0px;}";
    /* to hide things on printed page */
    html += "@media print { .unprintable { visibility: hidden;	} }";

    html += "</style>";

    //html += "<body onload='window.print()'>";
    html += "<body style='background-color: white;'>";

    return html;
}

function svg_page_header(page_width, page_height) {
    let svg = "<svg version='1.1' xmlns='http://www.w3.org/2000/svg'";
    svg += ` width='${page_width}mm' height='${page_height}mm' viewBox='0 0 ${page_width} ${page_height}'>`;

    return svg;
}

function printBarcodePage() {
    let frame = document.getElementById("print_frame");

    let window = frame.contentWindow;
    window.print();
}

function generate_barcode_page(target_id, page_layout, label_list, calibration) {
    let svg = svg_page_header(page_layout.page_width, page_layout.page_height);

    let c = calibration;

    console.log(calibration);

    svg += "<g id='barcode_page'";
    if (c !== undefined) {
	svg += ` transform='scale(${c.scalex}, ${c.scaley}),translate(${c.offsetx}, ${c.offsety})'`;
    }
    svg += '>';

    foreach_label(page_layout, function(column, row, count, xpos, ypos) {
	if (count >= label_list.length) { return; }

	let item = label_list[count];

	svg += svg_label(item.mode, item.label, item.tape_type, xpos, ypos, page_layout.label_borders);
    });

    svg += "</g>";
    svg += "</svg>";

    let html = html_page_header();
    html += svg;
    html += "</body>";
    html += "</html>";

    let frame = document.getElementById(target_id);

    setupPrintFrame(frame, page_layout.page_width, page_layout.page_height);

    let fwindow = frame.contentWindow;

    fwindow.document.open();
    fwindow.document.write(html);
    fwindow.document.close();
}

function setupPrintFrame(frame, page_width, page_height) {
    let dpi = 98;

    let dpr = window.devicePixelRatio;
    if (dpr !== undefined) {
	dpi = dpi*dpr;
    }

    let ppmm = dpi/25.4;

    frame.width = page_width*ppmm;
    frame.height = page_height*ppmm;
}

function generate_calibration_page(target_id, page_layout, calibration) {
    let frame = document.getElementById(target_id);

    setupPrintFrame(frame, page_layout.page_width, page_layout.page_height);

    let svg = svg_page_header(page_layout.page_width, page_layout.page_height);

    svg += "<defs>";
    svg += "<marker id='endarrow' markerWidth='10' markerHeight='7' ";
    svg += "refX='10' refY='3.5' orient='auto'><polygon points='0 0, 10 3.5, 0 7' />";
    svg += "</marker>";

    svg += "<marker id='startarrow' markerWidth='10' markerHeight='7' ";
    svg += "refX='0' refY='3.5' orient='auto'><polygon points='10 0, 10 7, 0 3.5' />";
    svg += "</marker>";
    svg += "</defs>";

    svg += "<rect x='50' y='50' width='100' height='100' style='fill:none;stroke-width:0.05;stroke:rgb(0,0,0)'/>";

    let text_style = "style='font-weight:bold;font-size:4;font-family:sans-serif;'";

    svg += `<text x='10' y='99' ${text_style}>Sx = 50mm</text>`;
    svg += "<line x1='0' y1='100' x2='50' y2='100' stroke='#000' marker-end='url(#endarrow)' stroke-width='.25'/>";

    svg += `<text x='60' y='99' ${text_style}>Dx = 100mm</text>`;
    svg += "<line x1='50' y1='100' x2='150' y2='100' stroke='#000' marker-start='url(#startarrow)' marker-end='url(#endarrow)' stroke-width='.25'/>";

    svg += `<text x='142' y='10' ${text_style} writing-mode='tb'>Sy = 50mm</text>`;
    svg += "<line x1='140' y1='0' x2='140' y2='50' stroke='#000' marker-end='url(#endarrow)' stroke-width='.25'/>";

    svg += `<text x='142' y='60' ${text_style} writing-mode='tb'>Dy = 100mm</text>`;
    svg += "<line x1='140' y1='50' x2='140' y2='150' stroke='#000' marker-start='url(#startarrow)' marker-end='url(#endarrow)' stroke-width='.25'/>";

    let c = calibration;
    if (c !== undefined) {
	svg += `<rect x='50' y='50' width='100' height='100' style='fill:none;stroke-width:0.05;stroke:rgb(255,0,0)' `;
	svg += `transform='scale(${c.scalex}, ${c.scaley}),translate(${c.offsetx}, ${c.offsety})'/>`;
    }

    svg += "</svg>";

    let html = html_page_header();
    html += svg;
    html += "</body>";
    html += "</html>";

    let fwindow = frame.contentWindow;

    fwindow.document.open();
    fwindow.document.write(html);
    fwindow.document.close();
}
Ext.define('PrefixField', {
    extend: 'Ext.form.field.Text',
    alias: 'widget.prefixfield',

    maxLength: 6,
    allowBlank: false,

    maskRe: /([A-Za-z]+)$/,

    listeners: {
	change: function(field) {
	    field.setValue(field.getValue().toUpperCase());
	},
    },
});
Ext.define('LtoLabelStyle', {
    extend: 'Ext.form.field.ComboBox',
    alias: 'widget.ltoLabelStyle',

    editable: false,

    displayField: 'text',
    valueField: 'value',
    queryMode: 'local',

    store: {
	field: ['value', 'text'],
	data: [
	    { value: 'simple', text: "Simple" },
	    { value: 'color', text: 'Color (frames with color)' },
	    { value: 'frame', text: 'Frame (no color)' },
	    { value: 'placeholder', text: 'Placeholder (empty)' },
	],
    },
});
Ext.define('LtoTapeType', {
    extend: 'Ext.form.field.ComboBox',
    alias: 'widget.ltoTapeType',

    editable: false,

    displayField: 'text',
    valueField: 'value',
    queryMode: 'local',

    store: {
	field: ['value', 'text'],
	data: [
	    { value: 'L8', text: "LTO-8" },
	    { value: 'L7', text: "LTO-7" },
	    { value: 'L6', text: "LTO-6" },
	    { value: 'L5', text: "LTO-5" },
	    { value: 'L4', text: "LTO-4" },
	    { value: 'L3', text: "LTO-3" },
	    { value: 'CU', text: "Cleaning Unit" },
	],
    },
});
const paper_sizes = {
    a4: {
	comment: 'A4 (plain)',
	page_width: 210,
	page_height: 297,
    },
    letter: {
	comment: 'Letter (plain)',
	page_width: 215.9,
	page_height: 279.4,
    },
    avery3420: {
	fixed: true,
	comment: 'Avery Zweckform 3420',
	page_width: 210,
	page_height: 297,
	label_width: 70,
	label_height: 16.9,
	margin_left: 0,
	margin_top: 5,
	column_spacing: 0,
	row_spacing: 0,
    },
};

function paper_size_combo_data() {
    let data = [];

    for (let [key, value] of Object.entries(paper_sizes)) {
	data.push({ value: key, text: value.comment });
    }
    return data;
}

Ext.define('PaperSize', {
    extend: 'Ext.form.field.ComboBox',
    alias: 'widget.paperSize',

    editable: false,

    displayField: 'text',
    valueField: 'value',
    queryMode: 'local',

    store: {
	field: ['value', 'text'],
	data: paper_size_combo_data(),
    },
});
Ext.define('PageLayoutPanel', {
    extend: 'Ext.panel.Panel',
    alias: 'widget.pageLayoutPanel',

    layout: {
	type: 'hbox',
	align: 'stretch',
	pack: 'start',
    },

    getValues: function() {
	let me = this;

	let values = {};

	Ext.Array.each(me.query('[isFormField]'), function(field) {
	    if (field.isValid()) {
		let data = field.getSubmitData();
		Ext.Object.each(data, function(name, val) {
		    values[name] = val;
		});
	    }
	});

	let paper_size = values.paper_size || 'a4';

	let param = Ext.apply({}, paper_sizes[paper_size]);
	if (param === undefined) {
	    throw `unknown paper size ${paper_size}`;
	}

	param.paper_size = paper_size;

	Ext.Object.each(values, function(name, val) {
	    let parsed = parseFloat(val, 10);
	    param[name] = isNaN(parsed) ? val : parsed;
	});

	return param;
    },

    controller: {
	xclass: 'Ext.app.ViewController',

	control: {
	    'paperSize': {
		change: function(field, paper_size) {
		    let view = this.getView();
		    let defaults = paper_sizes[paper_size];

		    let names = [
			'label_width',
			'label_height',
			'margin_left',
			'margin_top',
			'column_spacing',
			'row_spacing',
		    ];
		    for (i = 0; i < names.length; i++) {
			let name = names[i];
			let f = view.down(`field[name=${name}]`);
			let v = defaults[name];
			if (v != undefined) {
			    f.setValue(v);
			    f.setDisabled(defaults.fixed);
			} else {
			    f.setDisabled(false);
			}
		    }
		},
	    },
	    'field': {
		change: function() {
		    let view = this.getView();
		    let param = view.getValues();
		    view.fireEvent("pagechanged", param);
		},
	    },
	},
    },

    items: [
	{
	    border: false,
	    layout: {
		type: 'vbox',
		align: 'stretch',
		pack: 'start',
	    },
	    items: [
		{
		    xtype: 'paperSize',
		    name: 'paper_size',
		    value: 'a4',
		    fieldLabel: 'Paper Size',
		},
		{
		    xtype: 'numberfield',
		    name: 'label_width',
		    fieldLabel: 'Label width',
		    minValue: 70,
		    allowBlank: false,
		    value: 70,
		},
		{
		    xtype: 'numberfield',
		    name: 'label_height',
		    fieldLabel: 'Label height',
		    minValue: 15,
		    allowBlank: false,
		    value: 17,
		},
		{
		    xtype: 'checkbox',
		    name: 'label_borders',
		    fieldLabel: 'Label borders',
		    value: true,
		    inputValue: true,
		},
	    ],
	},
	{
	    border: false,
	    margin: '0 0 0 10',
	    layout: {
		type: 'vbox',
		align: 'stretch',
		pack: 'start',
	    },
	    items: [
		{
		    xtype: 'numberfield',
		    name: 'margin_left',
		    fieldLabel: 'Left margin',
		    minValue: 0,
		    allowBlank: false,
		    value: 0,
		},
		{
		    xtype: 'numberfield',
		    name: 'margin_top',
		    fieldLabel: 'Top margin',
		    minValue: 0,
		    allowBlank: false,
		    value: 4,
		},
		{
		    xtype: 'numberfield',
		    name: 'column_spacing',
		    fieldLabel: 'Column spacing',
		    minValue: 0,
		    allowBlank: false,
		    value: 0,
		},
		{
		    xtype: 'numberfield',
		    name: 'row_spacing',
		    fieldLabel: 'Row spacing',
		    minValue: 0,
		    allowBlank: false,
		    value: 0,
		},
	    ],
	},
    ],

});
Ext.define('PageCalibration', {
    extend: 'Ext.panel.Panel',
    alias: 'widget.pageCalibration',

    layout: {
	type: 'hbox',
	align: 'stretch',
	pack: 'start',
    },

    getValues: function() {
	let me = this;

	let values = {};

	Ext.Array.each(me.query('[isFormField]'), function(field) {
	    if (field.isValid()) {
		let data = field.getSubmitData();
		Ext.Object.each(data, function(name, val) {
		    let parsed = parseFloat(val, 10);
		    values[name] = isNaN(parsed) ? val : parsed;
		});
	    }
	});

	if (values.d_x === undefined) { return; }
	if (values.d_y === undefined) { return; }
	if (values.s_x === undefined) { return; }
	if (values.s_y === undefined) { return; }

	scalex = 100/values.d_x;
	scaley = 100/values.d_y;

	let offsetx = ((50 - values.s_x) - (50*scalex - 50))/scalex;
	let offsety = ((50 - values.s_y) - (50*scaley - 50))/scaley;

	return {
	    scalex: scalex,
	    scaley: scaley,
	    offsetx: offsetx,
	    offsety: offsety,
	};
    },

    controller: {
	xclass: 'Ext.app.ViewController',

	control: {
	    'field': {
		change: function() {
		    let view = this.getView();
		    let param = view.getValues();
		    view.fireEvent("calibrationchanged", param);
		},
	    },
	},
    },

    items: [
	{
	    border: false,
	    layout: {
		type: 'vbox',
		align: 'stretch',
		pack: 'start',
	    },
	    items: [
		{
		    xtype: 'displayfield',
		    value: 'a4',
		    fieldLabel: 'Start Offset Sx (mm)',
		    labelWidth: 150,
		    value: 50,
		},
		{
		    xtype: 'displayfield',
		    value: 'a4',
		    fieldLabel: 'Length Dx (mm)',
		    labelWidth: 150,
		    value: 100,
		},
		{
		    xtype: 'displayfield',
		    value: 'a4',
		    fieldLabel: 'Start Offset Sy (mm)',
		    labelWidth: 150,
		    value: 50,
		},
		{
		    xtype: 'displayfield',
		    value: 'a4',
		    fieldLabel: 'Length Dy (mm)',
		    labelWidth: 150,
		    value: 100,
		},
	    ],
	},
	{
	    border: false,
	    margin: '0 0 0 20',
	    layout: {
		type: 'vbox',
		align: 'stretch',
		pack: 'start',
	    },
	    items: [
		{
		    xtype: 'numberfield',
		    value: 'a4',
		    name: 's_x',
		    fieldLabel: 'Meassured Start Offset Sx (mm)',
		    allowBlank: false,
		    labelWidth: 200,
		},
		{
		    xtype: 'numberfield',
		    value: 'a4',
		    name: 'd_x',
		    fieldLabel: 'Meassured Length Dx (mm)',
		    allowBlank: false,
		    labelWidth: 200,
		},
		{
		    xtype: 'numberfield',
		    value: 'a4',
		    name: 's_y',
		    fieldLabel: 'Meassured Start Offset Sy (mm)',
		    allowBlank: false,
		    labelWidth: 200,
		},
		{
		    xtype: 'numberfield',
		    value: 'a4',
		    name: 'd_y',
		    fieldLabel: 'Meassured Length Dy (mm)',
		    allowBlank: false,
		    labelWidth: 200,
		},
	    ],
	},
    ],
});
Ext.define('LabelList', {
    extend: 'Ext.grid.Panel',
    alias: 'widget.labelList',

    plugins: {
        ptype: 'cellediting',
        clicksToEdit: 1,
    },

    selModel: 'cellmodel',

    store: {
	field: [
	    'prefix',
	    'tape_type',
	    {
		type: 'integer',
		name: 'start',
	    },
	    {
		type: 'integer',
		name: 'end',
	    },
	],
	data: [],
    },

    listeners: {
	validateedit: function(editor, context) {
	    console.log(context.field);
	    console.log(context.value);
	    context.record.set(context.field, context.value);
	    context.record.commit();
	    return true;
	},
    },

    columns: [
	{
            text: 'Prefix',
            dataIndex: 'prefix',
	    flex: 1,
	    editor: {
		xtype: 'prefixfield',
		allowBlank: false,
	    },
	    renderer: function(value, metaData, record) {
		console.log(record);
		if (record.data.mode === 'placeholder') {
		    return "-";
		}
		return value;
	    },
	},
	{
            text: 'Type',
            dataIndex: 'tape_type',
	    flex: 1,
	    editor: {
		xtype: 'ltoTapeType',
		allowBlank: false,
	    },
	    renderer: function(value, metaData, record) {
		console.log(record);
		if (record.data.mode === 'placeholder') {
		    return "-";
		}
		return value;
	    },
	},
	{
            text: 'Mode',
            dataIndex: 'mode',
	    flex: 1,
	    editor: {
		xtype: 'ltoLabelStyle',
		allowBlank: false,
	    },
	},
	{
	    text: 'Start',
	    dataIndex: 'start',
	    flex: 1,
	    editor: {
		xtype: 'numberfield',
		allowBlank: false,
	    },
	},
	{
	    text: 'End',
	    dataIndex: 'end',
	    flex: 1,
	    editor: {
		xtype: 'numberfield',
	    },
	    renderer: function(value) {
		if (value === null || value === '' || value === undefined) {
		    return "Fill";
		}
		return value;
	    },
	},
	{
	    xtype: 'actioncolumn',
	    width: 75,
	    items: [
		{
		    tooltip: 'Move Up',
		    iconCls: 'fa fa-arrow-up',
		    handler: function(grid, rowIndex) {
			if (rowIndex < 1) { return; }
			let store = grid.getStore();
			let record = store.getAt(rowIndex);
			store.removeAt(rowIndex);
			store.insert(rowIndex - 1, record);
		    },
		},
		{
		    tooltip: 'Move Down',
		    iconCls: 'fa fa-arrow-down',
		    handler: function(grid, rowIndex) {
			let store = grid.getStore();
			if (rowIndex >= store.getCount()) { return; }
			let record = store.getAt(rowIndex);
			store.removeAt(rowIndex);
			store.insert(rowIndex + 1, record);
		    },
		},
		{
		    tooltip: 'Delete',
		    iconCls: 'fa fa-scissors',
		    //iconCls: 'fa critical fa-trash-o',
		    handler: function(grid, rowIndex) {
			grid.getStore().removeAt(rowIndex);
		    },
		},
	    ],
	},
    ],
});
Ext.define('LabelSetupPanel', {
    extend: 'Ext.panel.Panel',
    alias: 'widget.labelSetupPanel',

    layout: {
	type: 'hbox',
	align: 'stretch',
	pack: 'start',
    },

    getValues: function() {
	let me = this;

	let values = {};

	Ext.Array.each(me.query('[isFormField]'), function(field) {
	    let data = field.getSubmitData();
	    Ext.Object.each(data, function(name, val) {
		let parsed = parseInt(val, 10);
		values[name] = isNaN(parsed) ? val : parsed;
	    });
	});

	return values;
    },

    controller: {
	xclass: 'Ext.app.ViewController',

	init: function() {
	    let me = this;
	    let view = me.getView();
	    let list = view.down("labelList");
	    let store = list.getStore();
	    store.on('datachanged', function(store) {
		view.fireEvent("listchanged", store);
	    });
	    store.on('update', function(store) {
		view.fireEvent("listchanged", store);
	    });
	},

	onAdd: function() {
	    let list = this.lookupReference('label_list');
	    let view = this.getView();
	    let params = view.getValues();
	    list.getStore().add(params);
	},
    },

    items: [
	{
	    border: false,
	    layout: {
		type: 'vbox',
		align: 'stretch',
		pack: 'start',
	    },
	    items: [
		{
		    xtype: 'prefixfield',
		    name: 'prefix',
		    value: 'TEST',
		    fieldLabel: 'Prefix',
		},
		{
		    xtype: 'ltoTapeType',
		    name: 'tape_type',
		    fieldLabel: 'Type',
		    value: 'L8',
		},
		{
		    xtype: 'ltoLabelStyle',
		    name: 'mode',
		    fieldLabel: 'Mode',
		    value: 'color',
		},
		{
		    xtype: 'numberfield',
		    name: 'start',
		    fieldLabel: 'Start',
		    minValue: 0,
		    allowBlank: false,
		    value: 0,
		},
		{
		    xtype: 'numberfield',
		    name: 'end',
		    fieldLabel: 'End',
		    minValue: 0,
		    emptyText: 'Fill',
		},
		{
		    xtype: 'button',
		    text: 'Add',
		    handler: 'onAdd',
		},
	    ],
	},
	{
	    margin: "0 0 0 10",
	    xtype: 'labelList',
	    reference: 'label_list',
	    flex: 1,
	},
    ],
});
// for toolkit.js
function gettext(val) { return val; };

function draw_labels(target_id, label_list, page_layout, calibration) {
    let max_labels = compute_max_labels(page_layout);

    let count_fixed = 0;
    let count_fill = 0;

    for (i = 0; i < label_list.length; i++) {
	let item = label_list[i];
	if (item.end === null || item.end === '' || item.end === undefined) {
	    count_fill += 1;
	    continue;
	}
	if (item.end <= item.start) {
	    count_fixed += 1;
	    continue;
	}
	count_fixed += (item.end - item.start) + 1;
    }

    let rest = max_labels - count_fixed;
    let fill_size = 1;
    if (rest >= count_fill) {
	fill_size = Math.floor(rest/count_fill);
    }

    let list = [];

    let count_fill_2 = 0;

    for (i = 0; i < label_list.length; i++) {
	let item = label_list[i];
	let count;
	if (item.end === null || item.end === '' || item.end === undefined) {
	    count_fill_2 += 1;
	    if (count_fill_2 === count_fill) {
		count = rest;
	    } else {
		count = fill_size;
	    }
	    rest -= count;
	} else if (item.end <= item.start) {
	    count = 1;
	} else {
	    count = (item.end - item.start) + 1;
	}

	for (j = 0; j < count; j++) {
	    let id = item.start + j;

	    if (item.prefix.length == 6) {
		list.push({
		    label: item.prefix,
		    tape_type: item.tape_type,
		    mode: item.mode,
		    id: id,
		});
		rest += count - j - 1;
		break;
	    } else {
		let pad_len = 6-item.prefix.length;
		let label = item.prefix + id.toString().padStart(pad_len, 0);

		if (label.length != 6) {
		    rest += count - j;
		    break;
		}

		list.push({
		    label: label,
		    tape_type: item.tape_type,
		    mode: item.mode,
		    id: id,
		});
	    }
	}
    }

    generate_barcode_page(target_id, page_layout, list, calibration);
}

Ext.define('MainView', {
    extend: 'Ext.container.Viewport',
    alias: 'widget.mainview',

    layout: {
	type: 'vbox',
	align: 'stretch',
	pack: 'start',
    },
    width: 800,

    controller: {
	xclass: 'Ext.app.ViewController',

	update_barcode_preview: function() {
	    let me = this;
	    let view = me.getView();
	    let list_view = view.down("labelList");

	    let store = list_view.getStore();
	    let label_list = [];
	    store.each((record) => {
		label_list.push(record.data);
	    });

	    let page_layout_view = view.down("pageLayoutPanel");
	    let page_layout = page_layout_view.getValues();

	    let calibration_view = view.down("pageCalibration");
	    let page_calibration = calibration_view.getValues();

	    draw_labels("print_frame", label_list, page_layout, page_calibration);
	},

	update_calibration_preview: function() {
	    let me = this;
	    let view = me.getView();
	    let page_layout_view = view.down("pageLayoutPanel");
	    let page_layout = page_layout_view.getValues();

	    let calibration_view = view.down("pageCalibration");
	    let page_calibration = calibration_view.getValues();
	    console.log(page_calibration);
	    generate_calibration_page('print_frame', page_layout, page_calibration);
	},

	control: {
	    labelSetupPanel: {
		listchanged: function(store) {
		    this.update_barcode_preview();
		},
		activate: function() {
		    this.update_barcode_preview();
		},
	    },
	    pageLayoutPanel: {
		pagechanged: function(layout) {
		    this.update_barcode_preview();
		},
		activate: function() {
		    this.update_barcode_preview();
		},
	    },
	    pageCalibration: {
		calibrationchanged: function() {
		    this.update_calibration_preview();
		},
		activate: function() {
		    this.update_calibration_preview();
		},
	    },
	},
    },

    items: [
	{
	    xtype: 'tabpanel',
	    items: [
		{
		    xtype: 'labelSetupPanel',
		    title: 'Proxmox LTO Barcode Label Generator',
		    bodyPadding: 10,
		},
		{
		    xtype: 'pageLayoutPanel',
		    title: 'Page Layout',
		    bodyPadding: 10,
		},
		{
		    xtype: 'pageCalibration',
		    title: 'Printer Calibration',
		    bodyPadding: 10,
		},
	    ],
	},
	{
	    xtype: 'panel',
	    layout: "center",
	    title: 'Print Preview',
	    bodyStyle: "background-color: grey;",
	    bodyPadding: 10,
	    html: '<center><iframe id="print_frame" frameBorder="0"></iframe></center>',
	    border: false,
	    flex: 1,
	    scrollable: true,
	    tools: [{
		type: 'print',
		tooltip: 'Open Print Dialog',
		handler: function(event, toolEl, panelHeader) {
		    printBarcodePage();
		},
	    }],
	},
    ],
});

Ext.onReady(function() {
    Ext.create('MainView', {
	renderTo: Ext.getBody(),
    });
});
